#!/usr/bin/python
# -*- coding: utf-8 -*-
import os
import json
import shutil
import hashlib
import binascii
import gzip
try:
    from cStringIO import StringIO
except ImportError:
    from StringIO import StringIO
import urllib2
from urllib import urlencode
from urlparse import urljoin

from utils import fetch_url


_opener_registered = False


def ensure_poster():
    global _opener_registered
    if not _opener_registered:
        from poster.streaminghttp import register_openers
        register_openers()
        _opener_registered = True


def md5_checksum(filepath, start=0, length=0):
    with open(filepath, 'rb') as fp:
        m = hashlib.md5()
        if start:
            fp.seek(start)

        bytes_read = 0
        while True:
            data = fp.read(8192)

            if not data:
                break

            if length > 0:
                bytes_read += len(data)
                if bytes_read >= length:
                    if bytes_read > length:
                        data = data[:length - bytes_read]
                    m.update(data)
                    break

            m.update(data)

        return m.hexdigest()


def crc32_checksum(filepath, start=0, length=0):
    crc = 0

    with open(filepath, 'rb') as fp:
        if start:
            fp.seek(start)

        bytes_read = 0
        while True:
            data = fp.read(8192)

            if not data:
                break

            if length > 0:
                bytes_read += len(data)
                if bytes_read >= length:
                    if bytes_read > length:
                        data = data[:length - bytes_read]
                    crc = binascii.crc32(data, crc)
                    break

            crc = binascii.crc32(data, crc)

    return crc & 0xFFFFFFFF

def is_torrent(filename):
    return filename.lower().endswith('.torrent')

def is_ed2k(url):
    return url.lower().startswith('ed2k://')

def is_magnet(url):
    return url.lower().startswith('magnet:')


class PCSApiError(Exception):

    def __init__(self, error_code, error_msg):
        self._error_code = error_code
        self._error_msg = error_msg

    def __str__(self):
        return 'PCS API Error %s: %s' % (self._error_code, self._error_msg)

class ClientApiError(Exception):

    def __init__(self, obj):
        self._obj = obj

    @property
    def obj(self):
        return self._obj

    def get_errmsg(self):
        obj = self._obj
        errno = obj.get('errno') or obj.get('error_code')
        if errno == 2:
            return '用户不存在'
        if errno == 4:
            return '密码错误'
        if errno == 16:
            return '该帐号已经超过登录次数，请过一段时间再试'
        if errno == -9999:
            return '加密验证出错，请稍候再试'

        return obj.get('errMsg') or obj.get('error_msg') or 'unknown error'

    def get_errno(self):
        obj = self._obj
        errno = obj.get('errno') or obj.get('error_code')
        return errno
    
    def __str__(self):
        obj = self._obj
        return 'API Error %s: %s' % (obj.get('errno') or obj.get('error_code'),
                                     self.get_errmsg()
                                     ) + '\n' + repr(obj)


class ApiBase:
    _pcsapi_baseUrl = 'https://pcs.baidu.com/rest/2.0/'

    def _create_pcsurl(self, path, params):
        return '%s?%s&access_token=%s' % \
            (self._pcsapi_baseUrl + path, urlencode(
                params), self._access_token)

    def _fetch_pcsapi(self, path, params=None, data=None, headers={}):
        url = urljoin(self._pcsapi_baseUrl, path) + '?'
        if params:
            url += urlencode(params) + '&'
        url += 'access_token=' + self._access_token

        try:
            r = fetch_url(url, data, headers)
        except urllib2.HTTPError as e:
            try:
                error_content = e.read()
                if e.headers.get('content-encoding') == 'gzip':
                    error_content = gzip.GzipFile(fileobj=StringIO(
                        error_content), mode='rb').read()
                eo = json.loads(error_content)

            except:
                raise e
            else:
                raise PCSApiError(eo.get('error_code'), eo.get('error_msg'))

        return json.loads(r)
        # handle json and return , and error msg thrown


class PCSApiMixin():

    def __init__(self, access_token):
        self._access_token = access_token

    def get_quota(self):
        return self._fetch_pcsapi('quota', {'method': 'info'})

    def upload_file(self, localfile, remotepath, overwrite=None,
                    tmpfile=False):
        from poster.encode import multipart_encode

        params = {'method': 'upload', 
                  'app_id': '250528', 
                  'dir': remotepath, 
                  'filename': os.path.basename(localfile)}
        if tmpfile:
            params['type'] = 'tmpfile'
        elif overwrite is not None:
            params['ondup'] = overwrite

        ensure_poster()
        datagen, headers = multipart_encode({"file": open(localfile, "rb")})

        return self._fetch_pcsapi('pcs/file', params, datagen, headers)

    def create_super_file(self, remotepath, filename, md5list, overwrite=None):
        params = {'method': 'createsuperfile',
                  'path': os.path.join(remotepath, filename)}
        if overwrite is not None:
            params['ondup'] = overwrite and 'overwrite' or 'newcopy'
        data = {'param': json.dumps({'block_list': md5list})}

        return self._fetch_pcsapi('file', params, data)

    def get_download_url(self, remotefile):
        return self._create_pcsurl(
            'file', {'method': 'download', 'path': remotefile})

    def download(self, remotefile, localpath=None, newfile=None):
        url = self.get_download_url(remotefile)
        if localpath:
            localfile = os.path.join(localpath,
                                     newfile or os.path.basename(remotefile))
            with open(localfile, 'wb') as fp:
                r = urllib2.urlopen(url)
                shutil.copyfileobj(r, fp)
            # or urllib retrieve
            return True

        else:
            return fetch_url(url)

    def mkdir(self, remotepath):
        return self._fetch_pcsapi(
            'file', {'method': 'mkdir', 'path': remotepath})

    def get_meta(self, remotepath):
        return self._fetch_pcsapi(
            'file', {'method': 'meta', 'path': remotepath})

    def get_batch_meta(self, paths):
        params = {'method': 'meta'}
        data = {'param': json.dumps({'list': [{'path': p} for p in paths]})}
        return self._fetch_pcsapi('file', params, data)

    def list(self, remotepath, by=None, asc=None, start=0, end=0):
        params = {'method': 'list', 'path': remotepath}
        if by is not None:
            params['by'] = by
        if asc is not None:
            params['order'] = asc and 'asc' or 'desc'
        if end > start:
            params['limit'] = '%d-%d' % (start, end)

        return self._fetch_pcsapi('file', params)

    def move(self, from_, to):
        data = {'from': from_, 'to': to}
        return self._fetch_pcsapi('file', {'method': 'move'}, data)

    def batch_move(self, *args):
        assert len(args) > 0
        data = {'param': json.dumps({'list': [{
                                    'from': f, 'to': t} for f, t in args]})}
        return self._fetch_pcsapi('file', {'method': 'move'}, data)

    def copy(self, from_, to):
        data = {'from': from_, 'to': to}
        return self._fetch_pcsapi('file', {'method': 'copy'}, data)

    def batch_copy(self, *args):
        assert len(args) > 0
        data = {'param': json.dumps({'list': [{
                                    'from': f, 'to': t} for f, t in args]})}
        return self._fetch_pcsapi('file', {'method': 'copy'}, data)

    def delete(self, path):
        data = {'path': path}
        return self._fetch_pcsapi('file', {'method': 'delete'}, data)

    def batch_delete(self, *args):
        nargs = len(args)
        assert nargs > 0
        if nargs == 1 and isinstance(args[0], (tuple, list)):
            paths = args[0]
        else:
            paths = args
        data = {'param': json.dumps({'list': [{'path': p} for p in paths]})}
        return self._fetch_pcsapi('file', {'method': 'delete'}, data)

    def search(self, path, keyword, recursive=False):
        params = {'method': 'search', 'path': path, 'wd': keyword}
        if recursive:
            params['re'] = 1
        return self._fetch_pcsapi('file', params)

    def get_thumbnail_url(self, picfile, width=160, height=None, quality=100):
        params = {'method': 'generate', 'path': picfile,
                  'width': width, 'height': height or width}
        if 0 < quality < 100:
            params['quality'] = quality
        return self._create_pcsurl('thumbnail', params)

    def diff(self, cursor='null'):
        params = {'method': 'diff', 'cursor': cursor}
        return self._fetch_pcsapi('file', params)

    def get_transcode_url(self, videofile, trans_type='M3U8_AUTO_720'):
        '''
            M3U8_320_240、M3U8_480_224、M3U8_480_360、M3U8_640_480、M3U8_854_480
            M3U8_AUTO_360、M3U8_AUTO_240、M3U8_AUTO_480、M3U8_AUTO_720
        '''
        params = {'method': 'streaming', 'path': videofile, 'type': trans_type}
        return self._create_pcsurl('file', params)

    def get_stream_file_list(self, stream_type='video', start=0, limit=1000,
                             filter_path=None):
        assert stream_type in ('video', 'audio', 'image', 'doc')

        params = {'method': 'list', 'type': stream_type}
        if start > 0:
            params['start'] = start
        if 0 < limit < 1000:
            params['limit'] = limit
        if filter_path:
            params['filter_path'] = filter_path
        return self._fetch_pcsapi('stream', params)

    def get_stream_url(self, remotefile):
        return self._create_pcsurl(
            'stream', {'method': 'download', 'path': remotefile})

    def _rapid_upload(self, fullpath, content_length, content_md5, slice_md5,
                      content_crc32, overwrite=None):
        params = {
            'method': 'rapidupload',
            'path': fullpath,
            'content-length': content_length,
            'content-md5': content_md5,
            'slice-md5': slice_md5,
            'content-crc32': content_crc32
        }
        if overwrite is not None:
            params['ondup'] = overwrite and 'overwrite' or 'newcopy'

        return self._fetch_pcsapi('file', params, '')

    def rapid_upload(self, localfile, path, newfile=None, overwrite=None):
        fullpath = os.path.join(path, newfile or os.path.basename(localfile))
        content_length = os.path.getsize(localfile)
        assert content_length > 256 * 1024

        content_md5 = md5_checksum(localfile)
        slice_md5 = md5_checksum(localfile, 0, 256 * 1024)
        content_crc32 = crc32_checksum(localfile)

        return self._rapid_upload(fullpath, content_length, content_md5,
                                  slice_md5, content_crc32, overwrite)


    def query_sinfo(self, source_path):
        params = {
            'method': 'query_sinfo',
            'source_path': source_path,
            'app_id': '250528',
            'type': '2',
            #'web': '1'
        }
        return self._fetch_pcsapi('services/cloud_dl', params, '')
    
    def query_shared_sinfo(self, source_path, uk, fs_id, shareid):
        params = {
            'method': 'query_sharedsinfo',
            'path': source_path,
            'uk': uk,
            'fs_id': fs_id,
            'sid': shareid,
            'app_id': '250528',
            'type': '2'
        }
        return self._fetch_pcsapi('services/cloud_dl', params, '')
    
    def query_magnetinfo(self, source_url):
        data = {
            'method': 'query_magnetinfo',
            'source_url': source_url,
            'app_id': '250528',
            'type': '4',
            #'web': '1'
        }
        return self._fetch_pcsapi('services/cloud_dl', '', data)
    
    def cloud_download(self, source_url, save_path, file_sha1, selected_idx, verifycode=None, vcode=None):
        params = {
            #'channel': 'chunlei',
            #'clienttype': '0',
            #'web': '1'
        }
        data = {
            'method': 'add_task',
            'save_path': save_path,
            'app_id': '250528',
            'file_sha1': file_sha1,
            'selected_idx': selected_idx,
            #'task_from': '1',
            #'t': '1393487413105',
        }
        if is_torrent(source_url):
            data['source_path'] = source_url
            data['type'] = '2'
        elif is_ed2k(source_url):
            data['source_url'] = source_url
            data['type'] = '3'
        elif is_magnet(source_url):
            data['source_url'] = source_url
            data['type'] = '4'

        if verifycode:
            data['input'] = verifycode
        if vcode:
            data['vcode'] = vcode
        
        return self._fetch_pcsapi('services/cloud_dl', params, data)
    
    def try_cloud_download(self, source_url, save_path, file_sha1, selected_idx, vcode_recognizer=None):
        vcode, verifycode = '', ''

        while True:
            try:
                return self.cloud_download(source_url, save_path, file_sha1, selected_idx, verifycode, vcode)
            except ClientApiError as e:
                r = e.obj
                error_code = r.get('error_code')
                if error_code == -19:
                    vcode = r['vcode']
                    img = r['img']
                    verifycode = vcode_recognizer(img)
                    if not verifycode:
                        return
                else:
                    raise

    def query_task(self, task_ids, op_type=1, expires=None):
        if not isinstance(task_ids, basestring):
            task_ids = ','.join(id for id in task_ids)

        params = {
            'method': 'query_task',
            'task_ids': task_ids,
            'op_type': op_type,
            'app_id': '250528'
        }

        if expires:
            params['expires'] = expires

        return self._fetch_pcsapi('services/cloud_dl', params, '')

    def list_task(self, expires=None, start=0, limit=10, asc=False,
                  source_url=None, save_path=None, create_time=None,
                  status=None, need_task_info=True):

        params = {
            'method': 'list_task',
        }

        if expires:
            params['expires'] = expires
        if start != 0:
            params['start'] = start
        if limit != 10:
            params['limit'] = limit
        if asc:
            params['asc'] = 1
        if source_url:
            params['source_url'] = source_url
        if save_path:
            params['save_path'] = save_path
        if create_time:
            params['create_time'] = create_time
        if status:
            params['status'] = status
        if not need_task_info:
            params['need_task_info'] = 0

        return self._fetch_pcsapi('services/cloud_dl', params, '')

    def cancel_task(self, task_id, expires=None):
        params = {
            'method': 'cancel_task',
            'task_id': task_id,
            'app_id': '250528'
        }

        if expires:
            params['expires'] = expires

        return self._fetch_pcsapi('services/cloud_dl', params, '')
    
    def delete_task(self, task_id, expires=None):
        params = {
            'method': 'delete_task',
            'task_id': task_id,
            'app_id': '250528'
        }

        if expires:
            params['expires'] = expires

        return self._fetch_pcsapi('services/cloud_dl', params, '')

    def list_cycle(self, start=0, limit=1000):
        params = {'method': 'listrecycle'}
        if start != 0:
            params['start'] = start
        if limit != 1000:
            params['limit'] = limit
        return self._fetch_pcsapi('file', params)

    def restore_from_cycle(self, *args):
        assert args > 0
        if len(args) == 1:
            return self._restore_from_cycle(args[0])

        return self._batch_restore_from_cycle(args)

    def _restore_from_cycle(self, fs_id):
        params = {'method': 'restore', 'fs_id': fs_id}

        return self._fetch_pcsapi('file', params, '')

    def _batch_restore_from_cycle(self, fs_ids):
        params = {
            'method': 'restore',
            'param': json.dumps({'list': [{'fs_id': p} for p in fs_ids]})
        }

        return self._fetch_pcsapi('file', params, '')

    def empty_cycle(self):
        params = {'method': 'delete', 'type': 'recycle'}
        return self._fetch_pcsapi('file', params, '')

    # def create_share_url(self, remotefile):
    #     params = {
    #         'method': 'create',
    #         'type': 'public',
    #         'path': remotefile
    #     }

    #     return self._fetch_pcsapi('share', params, '')


class PCSApi(PCSApiMixin, ApiBase):
    pass
